import xml.etree.ElementTree as ET

from dataclasses import dataclass, field
from typing import List


@dataclass
class BibliographicItemVersion:
    revision_date: str = None
    draft: List[str] = field(default_factory=list)

    def to_xml(self, parent=None):
        name = "version"
        node = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)
        if self.revision_date:
            ET.SubElement(node, "revision-date").text = self.revision_date
        for d in self.draft:
            ET.SubElement(node, "draft").text = d
        return node

    def to_asciibib(self, prefix=""):
        pref = f"{prefix}." if prefix else prefix
        out = []
        if self.revision_date:
            out.append(f"{pref}version.revision_date:: {self.revision_date}")
        for d in self.draft:
            out.append(f"{pref}version.draft:: {d}")
        return "\n".join(out)



import pytest
import inspect

import dataclasses
import xml.etree.ElementTree as ET

from relaton_bib import BiblioNote, BiblioNoteCollection


@pytest.fixture
def biblio_note():
    return BiblioNote(type="type", content="content")


@pytest.fixture
def bibnote_collection():
    return BiblioNoteCollection([
        BiblioNote(type="type1", content="content1", language=["en"]),
        BiblioNote(type="тип2", content="содержимое2", language=["ru"])])


def test_note_to_xml(biblio_note):
    host = ET.Element("host")
    result = biblio_note.to_xml(host)

    assert result.attrib["type"] == "type"
    assert host.find("./note").text == "content"


def test_note_to_asciibib(biblio_note):
    result = biblio_note.to_asciibib(prefix="test")

    assert result == inspect.cleandoc(
        """test.biblionote.type:: type
           test.biblionote.content:: content""")


def test_note_hash(biblio_note):
    result = dataclasses.asdict(biblio_note)

    assert result["type"] == "type"
    assert result["content"] == "content"


def test_note_collecion_to_xml(bibnote_collection):
    host = ET.Element("host")
    result = bibnote_collection.to_xml(host, {"lang": "en"})

    assert result is not None
    assert len(host.findall("./note")) == 1
    assert host.find("./note").text == "content1"
    assert host.find("./note").attrib["language"] == "en"


def test_note_collecion_to_xml_lang_not_found(bibnote_collection):
    host = ET.Element("host")
    result = bibnote_collection.to_xml(host, {"lang": "xx"})

    assert result is not None
    assert len(host.findall("./note")) == 2


def test_note_collecion_hash(bibnote_collection):
    result = dataclasses.asdict(bibnote_collection)

    assert result["array"][0]["type"] == "type1"
    assert result["array"][0]["content"] == "content1"
    assert result["array"][1] is not None



from __future__ import annotations
from dataclasses import dataclass
from typing import List

import datetime
import re
import xml.etree.ElementTree as ET

from .contribution_info import ContributionInfo
from .relaton_bib import to_ds_instance


@dataclass
class CopyrightAssociation:
    from_: datetime.date
    owner: List[ContributionInfo]

    to: datetime.date = None
    scope: str = None

    def __post_init__(self):
        if not self.owner:
            raise ValueError("at least one owner should exist.")

        self.owner = list(map(to_ds_instance(ContributionInfo), self.owner))

        if isinstance(self.from_, str) and re.match(r"\d{4}", self.from_):
            self.from_ = datetime.datetime.strptime(self.from_, "%Y")

        if isinstance(self.to, str) and self.to:
            self.to = datetime.datetime.strptime(self.to, "%Y")

    def _date_reset(self, d):
        return d.replace(month=1, day=1)

    def to_xml(self, parent, opts={}):
        name = "copyright"
        result = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)
        ET.SubElement(result, "from").text = str(self.from_.year) \
            if self.from_ else "unknown"
        if self.to:
            ET.SubElement(result, "to").text = str(self.to.year)
        for o in self.owner:
            o.to_xml(ET.SubElement(result, "owner"), opts)
        if self.scope:
            ET.SubElement(result, "scope").text = self.scope

        return result

    def to_asciibib(self, prefix="", count=1):
        pref = f"{prefix}.copyright" if prefix else "copyright"
        out = [f"{pref}::"] if count > 1 else []
        for ow in self.owner:
            out.append(ow.to_asciibib(f"{pref}.owner", len(self.owner)))
        if self.from_:
            out.append(f"{pref}.from:: {self.from_.year}")
        if self.to:
            out.append(f"{pref}.to:: {self.to.year}")
        if self.scope:
            out.append(f"{pref}.scope:: {self.scope}")
        return "\n".join(out)



import dataclasses
import inspect
import pytest

import xml.etree.ElementTree as ET

from relaton_bib import TypedTitleString, \
    TypedTitleStringCollection, FormattedString


def test_missing_title_or_content():
    with pytest.raises(ValueError) as excinfo:
        TypedTitleString(type="type")
    assert "Argument title or content should be passed" in str(excinfo.value)


@pytest.fixture
def subject():
    return TypedTitleString(
        type=TypedTitleString.Type.MAIN,
        title=FormattedString(content="Title", format=None)
    )


def test_valid_instance(subject):
    assert type(subject) is TypedTitleString


def test_to_xml(subject):
    host = ET.Element("host")
    subject.to_xml(host)

    assert host.attrib["type"] == TypedTitleString.Type.MAIN
    assert host.text == "Title"


def test_to_hash(subject):
    result = dataclasses.asdict(subject)

    assert result["type"] == TypedTitleString.Type.MAIN
    assert result["title"]["content"] == "Title"


def test_to_asciibib(subject):
    result = subject.to_asciibib()

    assert result == inspect.cleandoc(
        """title.type:: main
           title.content:: Title""")


def test_from_string_empty():
    t = TypedTitleString.from_string("")
    assert len(t) == 2
    assert t[0].title.content == ""
    assert t[0].type == TypedTitleString.Type.TMAIN
    assert t[1].title.content == ""
    assert t[1].type == TypedTitleString.Type.MAIN


def test_with_main():
    t = TypedTitleString.from_string("Main")
    assert len(t) == 2
    assert t[0].title.content == "Main"
    assert t[0].type == TypedTitleString.Type.TMAIN
    assert t[1].title.content == "Main"
    assert t[1].type == TypedTitleString.Type.MAIN


def test_with_main_and_part():
    t = TypedTitleString.from_string("Main - Part 1:")
    assert len(t) == 3
    assert t[0].title.content == "Main"
    assert t[0].type == TypedTitleString.Type.TMAIN
    assert t[1].title.content == "Part 1:"
    assert t[1].type == TypedTitleString.Type.TPART
    assert t[2].title.content == "Main - Part 1:"
    assert t[2].type == TypedTitleString.Type.MAIN


def test_with_intro_and_main():
    t = TypedTitleString.from_string("Intro - Main")
    assert len(t) == 3
    assert t[0].title.content == "Intro"
    assert t[0].type == TypedTitleString.Type.TINTRO
    assert t[1].title.content == "Main"
    assert t[1].type == TypedTitleString.Type.TMAIN
    assert t[2].title.content == "Intro - Main"
    assert t[2].type == TypedTitleString.Type.MAIN


def test_with_intro_and_main_and_part():
    t = TypedTitleString.from_string("Intro - Main - Part 1:")
    assert len(t) == 4
    assert t[0].title.content == "Intro"
    assert t[0].type == TypedTitleString.Type.TINTRO
    assert t[1].title.content == "Main"
    assert t[1].type == TypedTitleString.Type.TMAIN
    assert t[2].title.content == "Part 1:"
    assert t[2].type == TypedTitleString.Type.TPART
    assert t[3].title.content == "Intro - Main - Part 1:"
    assert t[3].type == TypedTitleString.Type.MAIN


def test_with_extra_part():
    t = TypedTitleString.from_string("Intro - Main - Part 1: - Extra")
    assert len(t) == 4
    assert t[0].title.content == "Intro"
    assert t[0].type == TypedTitleString.Type.TINTRO
    assert t[1].title.content == "Main"
    assert t[1].type == TypedTitleString.Type.TMAIN
    assert t[2].title.content == "Part 1: -- Extra"
    assert t[2].type == TypedTitleString.Type.TPART
    assert t[3].title.content == "Intro - Main - Part 1: -- Extra"
    assert t[3].type == TypedTitleString.Type.MAIN


def test_filter_collection(subject):
    c = TypedTitleStringCollection([])
    assert len(c) == 0
    c.append(subject)
    assert len(c) == 1
    r = TypedTitleStringCollection(filter(lambda t: t.type == "some", c))
    assert len(r) == 0
    r = TypedTitleStringCollection(
        filter(lambda t: t.type == TypedTitleString.Type.MAIN, c))
    assert len(r) == 1



import dataclasses
import inspect
import pytest

import xml.etree.ElementTree as ET

from relaton_bib import EditorialGroup, WorkGroup, TechnicalCommittee


@pytest.fixture
def subject():
    return EditorialGroup([
        TechnicalCommittee(WorkGroup(name="g1", number=1, type="work")),
        TechnicalCommittee(WorkGroup(name="g2", number=2)),
        TechnicalCommittee(WorkGroup(name="g3")),
    ])


def test_to_xml(subject):
    host = ET.Element("host")
    subject.to_xml(host)

    assert len(host.findall("./editorialgroup/technical-committee")) == 3


def test_to_asciibib(subject):
    result = subject.to_asciibib()

    assert result == inspect.cleandoc(
        """editorialgroup.technical_committee::
           editorialgroup.technical_committee.name:: g1
           editorialgroup.technical_committee.number:: 1
           editorialgroup.technical_committee.type:: work
           editorialgroup.technical_committee::
           editorialgroup.technical_committee.name:: g2
           editorialgroup.technical_committee.number:: 2
           editorialgroup.technical_committee::
           editorialgroup.technical_committee.name:: g3""")


def test_to_asciibib_with_pref(subject):
    result = subject.to_asciibib(prefix="g")

    assert result == inspect.cleandoc(
        """g.editorialgroup.technical_committee::
           g.editorialgroup.technical_committee.name:: g1
           g.editorialgroup.technical_committee.number:: 1
           g.editorialgroup.technical_committee.type:: work
           g.editorialgroup.technical_committee::
           g.editorialgroup.technical_committee.name:: g2
           g.editorialgroup.technical_committee.number:: 2
           g.editorialgroup.technical_committee::
           g.editorialgroup.technical_committee.name:: g3""")


def test_hash(subject):
    result = dataclasses.asdict(subject)

    assert result["technical_committee"][0]["workgroup"]["name"] == "g1"
    assert result["technical_committee"][0]["workgroup"]["number"] == 1
    assert result["technical_committee"][0]["workgroup"]["type"] == "work"
    assert result["technical_committee"][1]["workgroup"]["name"] == "g2"
    assert result["technical_committee"][1]["workgroup"]["number"] == 2
    assert result["technical_committee"][2]["workgroup"]["name"] == "g3"



import dataclasses
import inspect
import pytest

import xml.etree.ElementTree as ET

from relaton_bib import DocumentStatus


def test_create_with_hash():
    ds = DocumentStatus(stage={"value": "30", "abbreviation": "CD"})
    assert type(ds.stage) is DocumentStatus.Stage
    assert ds.stage.value == "30"
    assert ds.stage.abbreviation == "CD"


@pytest.fixture
def subject():
    return DocumentStatus(
        stage=DocumentStatus.Stage(value="30", abbreviation="CD"),
        substage=DocumentStatus.Stage(value="90", abbreviation="DVD"),
        iteration="5")


def test_to_xml(subject):
    host = ET.Element("host")
    result = subject.to_xml(host)

    assert host.find("./status") is not None
    assert result.find("./stage").text == "30"
    assert result.find("./stage").attrib["abbreviation"] == "CD"
    assert result.find("./substage").text == "90"
    assert result.find("./substage").attrib["abbreviation"] == "DVD"
    assert result.find("./iteration").text == "5"


def test_to_asciibib(subject):
    result = subject.to_asciibib()

    assert result == inspect.cleandoc(
        """docstatus.stage:: 30
           docstatus.substage:: 90
           docstatus.iteration:: 5""")


def test_to_asciibib_with_pref(subject):
    result = subject.to_asciibib(prefix="uh")

    assert result == inspect.cleandoc(
        """uh.docstatus.stage:: 30
           uh.docstatus.substage:: 90
           uh.docstatus.iteration:: 5""")


def test_hash(subject):
    result = dataclasses.asdict(subject)

    assert result["iteration"] == "5"
    assert result["stage"]["value"] == "30"
    assert result["stage"]["abbreviation"] == "CD"
    assert result["substage"]["value"] == "90"
    assert result["substage"]["abbreviation"] == "DVD"



import datetime
from typing import List

import bibtexparser
from bibtexparser.bparser import BibTexParser
import iso639

from .bibliographic_date import BibliographicDate, BibliographicDateType
from .bibliographic_item import BibliographicItem, BibliographicItemType
from .bib_item_locality import BibItemLocality
from .biblio_note import BiblioNote
from .classification import Classification
from .contribution_info import ContributionInfo, ContributorRole, \
                               ContributorRoleType
from .document_identifier import DocumentIdentifier
from .document_relation import DocumentRelation
from .formatted_string import FormattedString
from .localized_string import LocalizedString
from .person import Person, FullName
from .place import Place
from .series import Series, SeriesType
from .typed_title_string import TypedTitleString, TypedTitleStringCollection
from .typed_uri import TypedUri
from .organization import Organization


def from_bibtex(bibtex: str) -> dict:
    # https://github.com/sciunto-org/python-bibtexparser/issues/280#issuecomment-932478235
    parser = BibTexParser(common_strings=True)
    bt = bibtexparser.loads(bibtex, parser)

    return {e["ID"]: BibliographicItem(
        id=e["ID"],
        docidentifier=_fetch_docid(e),
        fetched=_fetch_fetched(e),
        type=_fetch_type(e),
        title=_fetch_title(e),
        contributor=_fetch_contributor(e),
        date=_fetch_date(e),
        place=_fetch_place(e),
        biblionote=_fetch_note(e),
        relation=_fetch_relation(e),
        extent=_fetch_extent(e),
        edition=e.get("edition"),
        series=_fetch_series(e),
        link=_fetch_link(e),
        language=_fetch_language(e),
        classification=_fetch_classification(e),
        keyword=_fetch_keyword(e)) for e in bt.entries}


def _fetch_docid(bibtex: dict) \
        -> List[DocumentIdentifier]:
    docid = []

    for key in ["isbn", "lccn", "isbn"]:
        value = bibtex.get(key)
        if value:
            docid.append(DocumentIdentifier(id=value, type=key))

    return docid


def _fetch_fetched(bibtex: dict) -> datetime.datetime:
    # TODO replace strptime with
    # https://dateutil.readthedocs.io/en/stable/index.html
    # to support more formats
    timestamp = bibtex.get("timestamp")
    return datetime.datetime.strptime(timestamp, "%Y-%m-%d %H:%M:%S") \
        if timestamp else None


def _fetch_type(bibtex: dict) -> str:
    # todo move all string constants to enums
    t = bibtex.get("type")
    if t in ["mastersthesis", "phdthesis"]:
        return BibliographicItemType.THESIS.value
    elif t == "conference":
        return BibliographicItemType.INPROCEEDINGS.value
    elif t == "misc":
        return BibliographicItemType.STANDARD.value
    else:
        return str(t)


def _fetch_place(bibtex: dict) -> List[Place]:
    address = bibtex.get("address")
    return [Place(address)] if address else []


def _fetch_title(bibtex: dict) -> TypedTitleStringCollection:
    titles = []
    for key in ["title", "subtitle"]:
        value = bibtex.get(key)
        if value:
            titles.append(TypedTitleString(type="main", content=value))
    return TypedTitleStringCollection(titles)


def _fetch_contributor(bibtex: dict) -> List[ContributionInfo]:
    contributor = []

    for key in [ContributorRoleType.AUTHOR, ContributorRoleType.EDITOR]:
        for entity in _fetch_person(bibtex.get(key)):
            contributor.append(
                ContributionInfo(entity=entity,
                                 role=[ContributorRole(type=key)]))

    for key in [ContributorRoleType.PUBLISHER,
                ContributorRoleType.INSTITUTION,
                ContributorRoleType.ORGANIZATION,
                ContributorRoleType.SCHOOL]:
        value = bibtex.get(key)
        if value:
            is_publisher = key == "publisher"
            descr = [] if is_publisher \
                else [FormattedString(content="sponsor")]
            entity_cls = Person if is_publisher else Organization
            contributor.append(
                ContributionInfo(
                    entity=entity_cls(name=value),
                    role=[ContributorRole(type=key, description=descr)]))

    return contributor


def _fetch_person(person) -> List[Person]:
    if not person:
        return []

    result = []
    for name in person:
        parts = name.split(", ")
        surname = LocalizedString(parts[0])
        fname = parts[1].split(" ") if len(parts) > 1 else []
        forename = [LocalizedString(fn) for fn in fname]
        pname = FullName(surname=surname, forename=forename)
        result.append(Person(name=pname))

    return result


def _fetch_date(bibtex: dict) -> List[BibliographicDate]:
    date = []
    year = bibtex.get("year")
    if year:
        on = datetime.date(int(year), int(bibtex.get("month_numeric",  1)), 1)
        date.append(BibliographicDate(
            type=BibliographicDateType.PUBLISHED, on=on))

    urldate = bibtex.get("urldate")
    if urldate:
        on = datetime.datetime.strptime(urldate, "%Y-%m-%d").date()
        date.append(BibliographicDate(
            type=BibliographicDateType.ACCESSED, on=on))

    return date


def _fetch_note(bibtex: dict) -> List[BiblioNote]:
    types = ["annote", "howpublished", "comment", "note", "content"]
    mapping = {"note": None, "content": "tableOfContents"}
    return [BiblioNote(type=mapping.get(k, k), content=bibtex.get(k))
            for k in types if k in bibtex]


def _fetch_relation(bibtex: dict) -> TypedTitleStringCollection:
    booktitle = bibtex.get("booktitle")
    if not booktitle:
        return []

    ttl = TypedTitleString(type=TypedTitleString.Type.MAIN, content=booktitle)
    title = TypedTitleStringCollection([ttl])
    return DocumentRelation(type=DocumentRelation.Type.partOf,
                            bibitem=BibliographicItem(title=title))


def _fetch_extent(bibtex: dict) -> List[BibItemLocality]:
    result = []
    for key in ["chapter", "pages", "volume"]:
        value = bibtex.get(key)
        if value:
            t = key
            if key == "pages":
                t = "page"
                from_, to = value.split("-")
            else:
                from_, to = (value, None)

            result.append(BibItemLocality(t, from_, to))

    return result


def _fetch_series(bibtex: dict) -> List[Series]:
    result = []
    journal = bibtex.get("journal")
    if journal:
        result.append(Series(
            type=SeriesType.JOURNAL,
            title=TypedTitleString(content=journal),
            number=bibtex.get("number")))
    series = bibtex.get("series")
    if series:
        result.append(Series(
            title=TypedTitleString(content=series)))
    return result
    pass


def _fetch_link(bibtex: dict) -> List[TypedUri]:
    link = []

    for key in ["url", "doi", "file2"]:
        value = bibtex.get(key)
        if value:
            t = "file" if key == "file2" else key
            link.append(TypedUri(type=t, content=value))

    return link


def _fetch_language(bibtex: dict) -> List[str]:
    lang = bibtex.get("language")
    if not lang:
        return []

    return [iso639.languages.get(name=lang.capitalize()).alpha2]


def _fetch_classification(bibtex: dict) -> List[Classification]:
    cls = []

    for key in ["type", "keyword", "mendeley-tags"]:
        value = bibtex.get(key)
        if value:
            t = "mendeley" if key == "mendeley-tags" else key
            cls.append(Classification(type=t, value=value))

    return cls


def _fetch_keyword(bibtex: dict) -> List[str]:
    keywords = bibtex.get("keywords")
    return map(lambda s: LocalizedString(s), keywords.split(", ")) \
        if keywords else []



import dataclasses
import inspect
import pytest

import xml.etree.ElementTree as ET

from relaton_bib import Classification


@pytest.fixture
def subject():
    return Classification(type="type", value="value")


def test_to_xml(subject):
    host = ET.Element("host")
    result = subject.to_xml(host)

    assert result.attrib["type"] == "type"
    assert host.find("./classification").text == "value"


def test_to_asciibib(subject):
    result = subject.to_asciibib()

    assert result == inspect.cleandoc("""classification.type:: type
                                         classification.value:: value""")


def test_to_asciibib_with_pref(subject):
    result = subject.to_asciibib(prefix="test")

    assert result == inspect.cleandoc("""test.classification.type:: type
                                         test.classification.value:: value""")


def test_hash(subject):
    result = dataclasses.asdict(subject)

    assert result["type"] == "type"
    assert result["value"] == "value"



from __future__ import annotations
from dataclasses import dataclass
from typing import List

import xml.etree.ElementTree as ET

from .technical_committee import TechnicalCommittee


@dataclass
class EditorialGroup:
    technical_committee: List[TechnicalCommittee]

    def to_xml(self, parent):
        name = "editorialgroup"
        result = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)
        for tc in self.technical_committee:
            tc.to_xml(result)
        return result

    def to_asciibib(self, prefix=""):
        pref = f"{prefix}.editorialgroup" if prefix else "editorialgroup"
        return "\n".join([tc.to_asciibib(pref, len(self.technical_committee))
                          for tc in self.technical_committee])

    @property
    def presence(self):
        return True



import os
import json
import yaml

from relaton_bib import from_dict


def test_create_bibitem_from_dict():
    file = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                        "examples",
                        "dict.json")

    with open(file) as json_file:
        reference = json.load(json_file)

        item = from_dict(reference)

        assert item is not None
        assert item.id == "IEC62531.2012Redline"
        assert item.abstract is not None


def test_create_bibitem_from_dict_yaml():
    file = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                        "examples",
                        "bib_item.yml")

    with open(file) as yaml_file:
        reference = yaml.safe_load(yaml_file)

        item = from_dict(reference)

        assert item is not None
        assert item.id == "ISOTC211"
        assert item.structuredidentifier is not None



import dataclasses
import inspect
import pytest

import xml.etree.ElementTree as ET

from relaton_bib import Affiliation, Organization, Person, \
    FullName, PersonIdentifier


def test_fillname_without_args():
    with pytest.raises(ValueError) as excinfo:
        FullName()
    assert "Should be given surname or completename" \
        in str(excinfo.value)


def test_invalid_personal_id_type():
    with pytest.raises(ValueError) as excinfo:
        Person(
            name=FullName(completename="John Lennon"),
            identifier=[PersonIdentifier("wrong_type", "value")])
    assert 'Invalid type. It should be "isni" or "uri".' \
        in str(excinfo.value)


def test_missing_name():
    with pytest.raises(ValueError) as excinfo:
        Person(identifier=[PersonIdentifier("isni", "value")])
    assert "missing name" \
        in str(excinfo.value)


@pytest.fixture
def subject():
    return Person(
        name=FullName(completename="John Lennon"),
        identifier=[PersonIdentifier("isni", "isni-value")],
        affiliation=[Affiliation(
            Organization(name="org1")
        )]
    )


def test_to_xml(subject):
    host = ET.Element("host")
    result = subject.to_xml(host)

    assert host.find("./person/name/completename").text == "John Lennon"
    assert result.find("./affiliation/organization/name").text == "org1"
    assert result.find("./identifier").text == "isni-value"
    assert result.find("./identifier").attrib["type"] == "isni"


def test_to_asciibib(subject):
    result = subject.to_asciibib()

    assert result == inspect.cleandoc(
        """name.completename:: John Lennon
           affiliation.organization.name:: org1
           type:: isni
           value:: isni-value""")


def test_to_asciibib_with_pref(subject):
    result = subject.to_asciibib(prefix="t")

    assert result == inspect.cleandoc(
        """t.name.completename:: John Lennon
           t.affiliation.organization.name:: org1
           t.type:: isni
           t.value:: isni-value""")


def test_hash(subject):
    result = dataclasses.asdict(subject)

    assert result["identifier"][0]["type"] == "isni"
    assert result["name"]["completename"]["content"] == "John Lennon"



import datetime
import dataclasses
import re

from typing import Dict, Union, Type, Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from .localized_string import LocalizedString


class RequestError(Exception):
    pass


def parse_date(date, str_res=True):
    """Convert date from string to date or string
    """

    if date is datetime.datetime:
        return date

    cases = [
        # February 2012
        (r"(?P<date>\w+\s\d{4})", "%B %Y", "%Y-%m"),
        # February 11, 2012
        (r"(?P<date>\w+\s\d{1,2},\s\d{4})", "%B %d, %Y", "%Y-%m-%d"),
        # 2012-02-11
        (r"(?P<date>\d{4}-\d{2}-\d{2})", "%Y-%m-%d", None),
        # 2012-02
        (r"(?P<date>\d{4}-\d{2})", "%Y-%m", None),
        # 2012
        (r"(?P<date>\d{4})", "%Y", None),
    ]

    for regexp, strp, strf in cases:
        m = re.match(regexp, str(date))
        if m:
            value = m.group("date")
            d = datetime.datetime.strptime(value, strp)
            if strf:
                return d.strftime(strf) if str_res else d
            else:
                return value if str_res else d.strftime(strp)


# TODO we don't need this for python probably
# @param array [Array]
# @return [Array<String>, String]
def single_element_array(array):
    if len(array) > 1:
        return map(lambda x: x if x is str else dataclasses.asdict(x), array)
    else:
        return array[0] if array[0] is str else dataclasses.asdict(array[0])


def lang_filter(target, opts={}):
    lang = opts.get("lang")
    filtered = list(filter(
        lambda t: t.language and lang in t.language,
        target))
    return filtered if filtered else target


def to_ds_instance(target: Union[Type, Callable], fail=False):
    def f(x):
        if isinstance(x, target):
            return x
        elif isinstance(x, dict):
            return target(**x)
        elif isinstance(x, str):
            return target(x)
        elif fail:
            ValueError(f"Unknown how to conver {type(x).__name__} to {target}")
        else:
            return x
    return f


def delegate(to, *methods):
    """https://stackoverflow.com/a/55563139/902217"""
    def dec(klass):
        def create_delegator(method):
            def delegator(self, *args, **kwargs):
                obj = getattr(self, to)
                m = getattr(obj, method)
                return m(*args, **kwargs)
            return delegator
        for m in methods:
            setattr(klass, m, create_delegator(m))
        return klass
    return dec


def dict_replace_key(d: Dict, keys_to_replace: Dict) -> Dict:
    for (old_key, new_key) in keys_to_replace.items():
        if old_key in d:
            d[new_key] = d.pop(old_key)
    return d



from dataclasses import dataclass, field
from enum import Enum

from .localized_string import LocalizedString


class FormattedStringFormat(str, Enum):
    TEXT_PLAIN = "text/plain"
    TEXT_HTML = "text/html"
    TEXT_MARKDOWN = "text/markdown"
    APPL_DOCBOOK_XML = "application/docbook+xml"
    APPL_TEI_XML = "application/tei+xml"
    APPL_X_ASCIIDOC = "text/x-asciidoc"
    APPL_X_ISODOC_XML = "application/x-isodoc+xml"


@dataclass(frozen=True)
class FormattedString(LocalizedString):
    format: str = field(default=FormattedStringFormat.TEXT_PLAIN.value)

    def __post_init__(self):
        super().__post_init__()
        if isinstance(self.format, FormattedStringFormat):
            object.__setattr__(self, "format", self.format.value)

    def to_xml(self, parent):
        if self.format:
            parent.attrib["format"] = self.format
        return super().to_xml(parent)

    def to_asciibib(self, prefix="", count=1, has_attrs=False):
        has_attrs = has_attrs or self.format
        pref = f"{prefix}." if prefix else prefix
        out = [super().to_asciibib(prefix, count, has_attrs)]
        if self.format:
            out.append(f"{pref}format:: {self.format}")
        return "\n".join(out)



from relaton_bib import BibliographicDate, BibliographicDateType


def test_nov_2014():
    date = BibliographicDate(type=BibliographicDateType.PUBLISHED.value,
                             on="November 2014")
    assert date.on == "2014-11"


def test_nov_2014_to_xml():
    date = BibliographicDate(type=BibliographicDateType.PUBLISHED.value,
                             on="November 2014")

    result = date.to_xml(None, {})

    assert result.attrib["type"] == BibliographicDateType.PUBLISHED.value
    assert result.find("./on").text == "2014-11"


def test_nov_2014_full_date():
    date = BibliographicDate(type=BibliographicDateType.ISSUED.value,
                             on="November 2014")

    result = date.to_xml(None, {"date_format": "full"})

    assert result.attrib["type"] == BibliographicDateType.ISSUED.value
    assert result.find("./on").text == "2014-11-01"


def test_from_to_xml():
    date = BibliographicDate(type=BibliographicDateType.COPIED.value,
                             from_="2014-11", to="2015-12")
    result = date.to_xml(None, {})

    assert result.attrib["type"] == BibliographicDateType.COPIED.value
    assert result.find("./from").text == "2014-11"
    assert result.find("./to").text == "2015-12"


def test_from_to_short_xml():
    date = BibliographicDate(type=BibliographicDateType.ADAPTED.value,
                             from_="2014-11", to="2015-12")

    result = date.to_xml(None, {"date_format": "short"})

    assert result.attrib["type"] == BibliographicDateType.ADAPTED.value
    assert result.find("./from").text == "2014-11"
    assert result.find("./to").text == "2015-12"


def test_on_format_year_only():
    date = BibliographicDate(type=BibliographicDateType.ACCESSED.value,
                             on="2014")

    result = date.to_xml(None, {"date_format": "full"})

    assert result.attrib["type"] == BibliographicDateType.ACCESSED.value
    assert result.find("./on").text == "2014-01-01"


# FIXME
# def test_date_not_matched_any_patterns():
#     date = BibliographicDate(type=BibliographicDateType.ACCESSED.value, on="")
#     item.instance_variable_set :@on, "Nov 2020"
#     expect(item.on(:month)).to eq "Nov 2020"



import dataclasses
import inspect
import logging
import xml.etree.ElementTree as ET

from relaton_bib import BibItemLocality, BibItemLocalityType


def test_to_xml_all_props():
    bib_item = BibItemLocality(
        type=BibItemLocalityType.SECTION.value,
        reference_from="from",
        reference_to="to")

    host = ET.Element("host")
    result = bib_item.to_xml(host)

    assert result.attrib["type"] == BibItemLocalityType.SECTION.value
    assert host.find("./referenceFrom").text == "from"
    assert host.find("./referenceTo").text == "to"


def test_to_xml_only_required_props():
    bib_item = BibItemLocality(
        type=BibItemLocalityType.CLAUSE.value,
        reference_from="from")

    host = ET.Element("host")
    result = bib_item.to_xml(host)

    assert result.attrib["type"] == BibItemLocalityType.CLAUSE.value
    assert host.find("./referenceFrom").text == "from"
    assert host.find("./referenceTo") is None


def test_invalid_locality_type(caplog):
    invalid = "invalid"

    with caplog.at_level(logging.WARNING):
        BibItemLocality(type=invalid, reference_from="from")

    assert f"[relaton-bib] invalid locality type: {invalid}" in caplog.text


def test_custom_locality_type(caplog):
    custom = "locality:custom"

    with caplog.at_level(logging.WARNING):
        BibItemLocality(type=custom, reference_from="from")

    assert f"[relaton-bib] invalid locality type: {custom}" not in caplog.text


def test_to_asciibib_all_props():
    bib_item = BibItemLocality(
        type=BibItemLocalityType.SECTION.value,
        reference_from="from",
        reference_to="to")

    result = bib_item.to_asciibib(prefix="test")

    assert result == inspect.cleandoc("""test.type:: section
                                      test.reference_from:: from
                                      test.reference_to:: to""")


def test_to_asciibib_only_required_props():
    bib_item = BibItemLocality(
        type=BibItemLocalityType.SECTION.value,
        reference_from="from")

    result = bib_item.to_asciibib(count=2)

    assert result == inspect.cleandoc("""::
                                      type:: section
                                      reference_from:: from""")


def test_hash_all_props():
    bib_item = BibItemLocality(
        type=BibItemLocalityType.SECTION.value,
        reference_from="from",
        reference_to="to")

    result = dataclasses.asdict(bib_item)

    assert result["type"] == BibItemLocalityType.SECTION.value
    assert result["reference_from"]["content"] == "from"
    assert result["reference_to"]["content"] == "to"


def test_hash_only_required_props():
    bib_item = BibItemLocality(
        type=BibItemLocalityType.SECTION.value,
        reference_from="from")

    result = dataclasses.asdict(bib_item)

    assert result["type"] == BibItemLocalityType.SECTION.value
    assert result["reference_from"]["content"] == "from"



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages


def empty_or_comment(x):
    return len(x) == 0 or x.strip().startswith("#")


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

with open('requirements.txt') as reqs:
    requirements = filter(empty_or_comment, reqs.read().splitlines())

with open('requirements_dev.txt') as reqs:
    dev_requirements = filter(empty_or_comment, reqs.read().splitlines())

setup(
    author="Aliaksandr Babrykovich",
    author_email='abobrikovich@gmail.com',
    python_requires='>=3.8',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="""
        Python implementation of bibliography specification
        https://github.com/relaton/relaton-models#bibliography-uml-models
    """,
    install_requires=requirements,
    license="BSD license",
    long_description=readme + '\n\n' + history,
    long_description_content_type="text/x-rst",
    include_package_data=True,
    keywords='relaton_bib',
    name='relaton_bib',
    packages=find_packages(include=['relaton_bib', 'relaton_bib.*']),
    test_suite='tests',
    tests_require=dev_requirements,
    url='https://github.com/relaton/relaton-bib-py',
    version='0.2.0',
    zip_safe=False,
)



import dataclasses
import inspect
import pytest

import xml.etree.ElementTree as ET

from relaton_bib import Address, Contact, ContactType, Contributor


@pytest.fixture
def subject():
    return Contributor(
        uri="https://contributor.uri",
        contact=[
            # all props
            Address(
                street=["street", "1", "/2"],
                city="NY",
                country="USA",
                state="NY",
                postcode="10001"),
            # only required props
            Address(
                street=["vulica", "3", "/4"],
                city="MSQ",
                country="BY"),
            Contact(
                type=ContactType.EMAIL.value,
                value="test@test.com")
        ])


def test_url_eq_uri(subject):
    assert subject.uri == subject.url


def test_to_xml(subject):
    host = ET.Element("host")
    result = subject.to_xml(host)

    assert len(host.findall(f"./{ContactType.EMAIL.value}")) == 1
    assert host.find(f"./{ContactType.EMAIL.value}").text == "test@test.com"

    assert len(host.findall("./address")) == 2
    assert host.find("./address/state").text == "NY"
    assert host.find("./address/postcode").text == "10001"
    assert host.find("./address/street[1]").text == "street"
    assert host.find("./address[2]/street[1]").text == "vulica"
    assert host.find("./address/city").text == "NY"
    assert host.find("./address[2]/city").text == "MSQ"


def test_to_asciibib(subject):
    result = subject.to_asciibib()

    assert result == inspect.cleandoc(
        """url:: https://contributor.uri
           address::
           address.street:: street
           address.street:: 1
           address.street:: /2
           address.city:: NY
           address.state:: NY
           address.country:: USA
           address.postcode:: 10001
           address::
           address.street:: vulica
           address.street:: 3
           address.street:: /4
           address.city:: MSQ
           address.country:: BY
           contact.type:: email
           contact.value:: test@test.com""")


def test_to_asciibib_with_pref(subject):
    result = subject.to_asciibib(prefix="test")

    assert result == inspect.cleandoc(
        """test.url:: https://contributor.uri
           test.address::
           test.address.street:: street
           test.address.street:: 1
           test.address.street:: /2
           test.address.city:: NY
           test.address.state:: NY
           test.address.country:: USA
           test.address.postcode:: 10001
           test.address::
           test.address.street:: vulica
           test.address.street:: 3
           test.address.street:: /4
           test.address.city:: MSQ
           test.address.country:: BY
           test.contact.type:: email
           test.contact.value:: test@test.com""")


def test_hash(subject):
    result = dataclasses.asdict(subject)

    assert result["uri"] == "https://contributor.uri"
    assert len(result["contact"]) == 3



from __future__ import annotations
import datetime
import logging
import pytest
import re
import os
import xml.etree.ElementTree as ET

from lxml import etree

from . import elements_equal

from relaton_bib import BibliographicItem, BibliographicItemType
from relaton_bib import Address
from relaton_bib import Contact
from relaton_bib import Affiliation
from relaton_bib import TypedUri
from relaton_bib import DocumentIdentifier, DocumentIdType
from relaton_bib import CopyrightAssociation
from relaton_bib import FormattedString, FormattedStringFormat
from relaton_bib import ContributionInfo, ContributorRole
from relaton_bib import BibliographicDate, BibliographicDateType
from relaton_bib import Series, SeriesType
from relaton_bib import DocumentStatus
from relaton_bib import Organization, OrgIdentifier
from relaton_bib import LocalizedString
from relaton_bib import TypedTitleString, TypedTitleStringCollection
from relaton_bib import TechnicalCommittee
from relaton_bib import FormattedRef
from relaton_bib import Medium
from relaton_bib import Classification
from relaton_bib import Validity
from relaton_bib import BibItemLocality, Locality, BibItemLocalityType, \
    SourceLocalityStack, SourceLocality, LocalityStack
from relaton_bib import BiblioNote, BiblioNoteCollection
from relaton_bib import BibliographicItemVersion
from relaton_bib import Place
from relaton_bib import Person, FullName, PersonIdentifier
from relaton_bib import StructuredIdentifierCollection
from relaton_bib import EditorialGroup
from relaton_bib import ICS
from relaton_bib import DocRelationCollection
from relaton_bib import DocumentRelation
from relaton_bib import WorkGroup
from relaton_bib import StructuredIdentifier


@pytest.fixture
def title() -> TypedTitleStringCollection:
    result = TypedTitleString.from_string("Geographic information")
    result.append(TypedTitleString(content="Information géographique",
                                   language=["fr"],
                                   script=["Latn"]))
    return result


@pytest.fixture
def subject(title: TypedTitleStringCollection) -> BibliographicItem:
    return BibliographicItem(
        id="ISOTC211",
        title=title,
        fetched=datetime.date(2021, 8, 30),
        type=BibliographicItemType.STANDARD,
        doctype="document",
        subdoctype="subdocument",
        docidentifier=[
            DocumentIdentifier("TC211", DocumentIdType.ISO),
            DocumentIdentifier("ISBN", "isbn"),
            DocumentIdentifier("LCCN", "lccn"),
            DocumentIdentifier("ISSN", "issn", "series"),
            DocumentIdentifier("urn:iso:std:iso:123:stage-90.93:ed-3:en,fr",
                               "URN"),
            DocumentIdentifier("XYZ"),
            DocumentIdentifier("10.17487/rfc1149", "DOI"),
            DocumentIdentifier("draft-ietf-somewg-someprotocol-07",
                               "Internet-Draft")
        ],
        docnumber="123456",
        edition="1",
        language=["en", "fr"],
        script=["Latn"],
        version=BibliographicItemVersion(revision_date="2019-04-01",
                                         draft=["draft"]),
        biblionote=BiblioNoteCollection([
            BiblioNote(content="note"),
            BiblioNote(content="An note", type="annote"),
            BiblioNote(content="How published", type="howpublished"),
            BiblioNote(content="Comment", type="comment"),
            BiblioNote(content="Table Of Contents", type="tableOfContents"),

        ]),
        status=DocumentStatus(
            stage=DocumentStatus.Stage(value="30", abbreviation="CD"),
            substage=DocumentStatus.Stage(value="substage"),
            iteration="final"
        ),
        date=[
            BibliographicDate(
                type=BibliographicDateType.ISSUED, on="2014"),
            BibliographicDate(
                type=BibliographicDateType.PUBLISHED, on="2014-04"),
            BibliographicDate(
                type=BibliographicDateType.ACCESSED, on="2015-05-20")
        ],
        abstract=[
            FormattedString(
                content="ISO 19115-1:2014 defines the schema required for ...",
                language="en",
                script="Latn",
                format="text/plain"
            ),
            FormattedString(
                content="L'ISO 19115-1:2014 définit le schéma requis pour ...",
                language="fr",
                script="Latn",
                format="text/plain"
            )
        ],
        contributor=[
            ContributionInfo(
                role=[
                    ContributorRole(
                        type="publisher",
                        description=[FormattedString(content="Publisher role",
                                                     format=None)])
                ],
                entity=Organization(
                    name="International Organization for Standardization",
                    abbreviation="ISO",
                    subdivision=[LocalizedString("division")],
                    uri="www.iso.org")
            ),
            ContributionInfo(
                role=[ContributorRole(type="author")],
                entity=Person(
                    name=FullName(completename=LocalizedString(
                        content="A. Bierman",
                        language="en",
                        script="Latn")),
                    affiliation=[Affiliation(Organization(
                        name="IETF",
                        abbreviation="IETF",
                        identifier=[OrgIdentifier(
                            type="uri",
                            value="www.ietf.org")]
                    ))],
                    contact=[
                        Address(
                            street=["Street"],
                            city="City",
                            state="State",
                            country="Country",
                            postcode="123456"),
                        Contact(type="phone", value="223322")
                    ]
                )
            ),
            ContributionInfo(
                role=[
                    ContributorRole(
                        type="publisher",
                        description=[FormattedString(
                            content="Publisher description",
                            format=None)]),
                    ContributorRole(
                        type="editor",
                        description=[FormattedString(
                            content="Editor description", format=None)])
                ],
                entity=Organization(
                    name="IETF",
                    abbreviation="IETF",
                    identifier=[OrgIdentifier(
                        type="uri",
                        value="www.ietf.org")])
            ),
            ContributionInfo(
                role=[ContributorRole(type="author")],
                entity=Person(
                    name=FullName(
                        initial=[LocalizedString(
                            content="A.", language="en", script="Latn")],
                        surname=LocalizedString(
                            content="Bierman", language="en", script="Latn"),
                        forename=[LocalizedString(
                            content="Forename", language="en", script="Latn")],
                        addition=[LocalizedString(
                            content="Addition", language="en", script="Latn")],
                        prefix=[LocalizedString(
                            content="Prefix", language="en", script="Latn")]),
                    affiliation=[Affiliation(
                        organization=Organization(
                            name="IETF",
                            abbreviation="IETF"),
                        description=[FormattedString(
                            content="Description",
                            language="en",
                            format=None)]
                    )],
                    contact=[
                        Address(
                            street=["Street"],
                            city="City",
                            postcode="123456",
                            country="Country",
                            state="State"),
                        Contact(type="phone", value="223322")
                    ],
                    identifier=[PersonIdentifier(
                        type="uri",
                        value="www.person.com")]
                )
            ),
            ContributionInfo(
                entity=Organization(name="Institution"),
                role=[ContributorRole(
                    type="distributor",
                    description=[FormattedString(content="sponsor",
                                                 format=None)])],
            )
        ],
        copyright=[CopyrightAssociation(
            owner=[ContributionInfo(
                Organization(
                    name="International Organization for Standardization",
                    abbreviation="ISO",
                    uri="www.iso.org"))],
            from_="2014",
            to="2020",
            scope="Scope"
        )],
        link=[
            TypedUri(type="src",
                     content="https://www.iso.org/standard/53798.html"),
            TypedUri(type="obp",
                     content="https://www.iso.org/obp/ui/#!iso:std:53798:en"),
            TypedUri(
                type="rss",
                content="https://www.iso.org/contents/data/standard/05/37/"
                        "53798.detail.rss"),
            TypedUri(type="doi", content="http://standrd.org/doi-123"),
            TypedUri(type="file", content="file://path/file"),
        ],
        relation=DocRelationCollection([
            DocumentRelation(
                type=DocumentRelation.Type.updates,
                bibitem=BibliographicItem(
                    formattedref=FormattedRef(
                        content="ISO 19115:2003",
                        format=None)),
                locality=[
                    LocalityStack([
                        Locality(
                            type="section",
                            reference_from="Reference from")
                    ]),
                    LocalityStack([
                        Locality(type="chapter", reference_from="1"),
                        Locality(type="page", reference_from="2")
                    ])
                ],
                source_locality=[
                    SourceLocalityStack([
                        SourceLocality(type="volume", reference_from="2"),
                        SourceLocality(type="chapter", reference_from="3")

                    ])
                ]),
            DocumentRelation(
                type=DocumentRelation.Type.obsoletes,
                description=FormattedString(format="text/plain",
                                            content="supersedes"),
                bibitem=BibliographicItem(
                    type=BibliographicItemType.STANDARD,
                    formattedref=FormattedRef(
                        content="ISO 19115:2003/Cor 1:2006",
                        format=None))),
            DocumentRelation(
                type=DocumentRelation.Type.partOf,
                bibitem=BibliographicItem(
                    title=TypedTitleStringCollection(
                        [TypedTitleString(
                            type=TypedTitleString.Type.MAIN,
                            content="Book title")]))
            )
        ]),
        series=[
            Series(
                type=SeriesType.MAIN,
                title=TypedTitleString(
                    type="original",
                    content="ISO/IEC FDIS 10118-3",
                    language="en",
                    script="Latn",
                    format="text/plain"
                ),
                place="Serie's place",
                organization="Serie's organization",
                abbreviation=LocalizedString("ABVR"),
                from_="2009-02-01",
                to="2010-12-20",
                number="serie1234",
                partnumber="part5678"),
            Series(
                type="alt",
                formattedref=FormattedRef(
                    content="serieref",
                    language="en",
                    script="Latn")
            ),
            Series(
                type="journal",
                title=TypedTitleString(content="Journal"),
                number="7"
            ),
            Series(
                title=TypedTitleString(content=[
                    LocalizedString(content="Series",
                                    language="en",
                                    script="Latn"),
                    LocalizedString(content="Séries",
                                    language="fr",
                                    script="Latn")])
            ),
            Series(
                title=TypedTitleString(content="RFC"),
                number="4"
            )
        ],
        medium=Medium(
            form="medium form",
            size="medium size",
            scale="medium scale"),
        place=[
            Place("bib place"),
            Place(name="Geneva", uri="geneva.place", region="Switzelznd")
        ],
        extent=[
            BibItemLocality(
                type=BibItemLocalityType.SECTION,
                reference_from="Reference from",
                reference_to="Reference to"),
            BibItemLocality(
                type=BibItemLocalityType.CHAPTER,
                reference_from="4"),
            BibItemLocality(
                type=BibItemLocalityType.PAGE,
                reference_from="10",
                reference_to="20"),
            BibItemLocality(
                type=BibItemLocalityType.VOLUME,
                reference_from="1"),
        ],
        accesslocation=["accesslocation1", "accesslocation2"],
        classification=[
            Classification(type="type", value="value"),
            Classification("Keywords", "keyword"),
            Classification("Mendeley Tags", "mendeley")
        ],
        keyword=[LocalizedString("Keyword"), LocalizedString("Key Word")],
        license=["License"],
        validity=Validity(
            begins=datetime.datetime.strptime("2010-10-10 12:21",
                                              Validity.FORMAT),
            ends=datetime.datetime.strptime("2011-02-03 18:30",
                                            Validity.FORMAT),
            revision=datetime.datetime.strptime("2011-03-04 09:00",
                                                Validity.FORMAT),
        ),
        editorialgroup=EditorialGroup([
            TechnicalCommittee(
                WorkGroup(
                    name="Editorial group",
                    number=1,
                    type="Type",
                    identifier="Identifier",
                    prefix="Prefix"
                ))
        ]),
        ics=[ICS(code="01", text="First")],
        structuredidentifier=StructuredIdentifierCollection([
            StructuredIdentifier(
                type="type 1",
                agency=["agency 1", "agency 2"],
                class_="class 1",
                docnumber="123",
                partnumber="4",
                edition="1",
                version="2",
                supplementtype="type 2",
                supplementnumber="5",
                language="en",
                year="2020"),
            StructuredIdentifier(docnumber="456", agency=["agency 3"])
        ])
    )


def test_subject_created(subject: BibliographicItem):
    assert type(subject) == BibliographicItem


def test_titles(subject: BibliographicItem):
    assert type(subject.title) == TypedTitleStringCollection
    assert subject.title_for_lang("fr")[0].title.content == \
        "Information g\u00E9ographique"


def test_urls(subject: BibliographicItem):
    assert subject.url() == "https://www.iso.org/standard/53798.html"
    assert subject.url("rss") \
        == "https://www.iso.org/contents/data/standard/05/37/53798.detail.rss"


def test_shortref(subject: BibliographicItem):
    assert subject.shortref(subject.docidentifier[0]) == "TC211:2014"


def test_abstract(subject: BibliographicItem):
    assert isinstance(subject.abstract_for_lang("en"), FormattedString)


def test_to_most_recent_reference(subject: BibliographicItem):
    item = subject.to_most_recent_reference()
    assert item.relation[3].bibitem.structuredidentifier[0].year == "2020"
    assert item.structuredidentifier[0].year is None


def test_to_all_parts(subject: BibliographicItem):
    item = subject.to_all_parts()
    assert item != subject
    assert item.all_parts
    assert item.relation[-1].type == DocumentRelation.Type.instance
    assert next(
        (t for t in item.title if t.type == TypedTitleString.Type.TPART),
        None) is None
    assert next(
        (t.title.content for t in item.title
         if t.type == TypedTitleString.Type.MAIN),
        None) == "Geographic information"
    assert len(item.abstract) == 0
    assert next(
        (d for d in item.docidentifier
         if d.type != "Internet-Draft" and re.match(r"-\d", d.id)),
        None) is None
    assert len([d for d in item.docidentifier
                if "(all parts)" not in d.id]) == 1
    assert next(
        (d for d in item.docidentifier if re.match(r":[12]\d\d\d", d.id)),
        None) is None
    assert next(
        (si for si in item.structuredidentifier if si.partnumber),
        None) is None
    assert next(
        (si for si in item.structuredidentifier
            if re.match(r"-\d", si.docnumber)),
        None) is None
    assert next(
        (si for si in item.structuredidentifier
            if "(all parts)" not in si.docnumber),
        None) is None
    assert next(
        (si for si in item.structuredidentifier
            if re.match(r":[12]\d\d\d", si.docnumber)),
        None) is None


def test_bibitem_xml_string(subject: BibliographicItem):
    file = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                        "examples",
                        "bib_item.xml")
    tree = ET.parse(file)
    reference = tree.getroot()

    subject_xml = subject.to_xml()
    assert elements_equal(reference, subject_xml)

    tree = etree.parse(file)
    relaxng_document = etree.parse(
        os.path.join(os.path.dirname(os.path.abspath(__file__)),
                     "../grammars/biblio.rng"))
    xml_document = etree.fromstring(ET.tostring(subject_xml))
    relaxng_processor = etree.RelaxNG(relaxng_document)
    assert relaxng_processor.validate(xml_document)


def test_bibdata_xml_string(subject: BibliographicItem):
    file = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                        "examples",
                        "bibdata_item.xml")
    tree = ET.parse(file)
    reference = tree.getroot()

    subject_xml = subject.to_xml(opts={"bibdata": True})
    assert elements_equal(reference, subject_xml)

    tree = etree.parse(file)
    relaxng_document = etree.parse(
        os.path.join(os.path.dirname(os.path.abspath(__file__)),
                     "examples/isobib.rng"))
    xml_document = etree.fromstring(ET.tostring(subject_xml))
    relaxng_processor = etree.RelaxNG(relaxng_document)
    assert relaxng_processor.validate(xml_document)


def test_render_only_fr_lang_tagged_string(subject: BibliographicItem):
    file = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                        "examples",
                        "bibdata_item_fr.xml")
    tree = ET.parse(file)
    reference = tree.getroot()

    subject_xml = subject.to_xml(opts={"bibdata": True, "lang": "fr"})
    assert elements_equal(reference, subject_xml)


def test_render_addition_elements(subject: BibliographicItem):
    def block(node, _):
        ET.SubElement(node, "element").text = "test"
    xml = subject.to_xml(opts={"lambda": block})
    assert b"<element>test</element>" in ET.tostring(xml)


def test_add_note_to_xml(subject: BibliographicItem):
    opts = {"note": [{"text": "Note", "type": "note"}]}
    xml = subject.to_xml(opts=opts)
    assert b"<note format=\"text/plain\" type=\"note\">Note</note>" \
        in ET.tostring(xml)


def test_to_bibtex_standard(subject: BibliographicItem):
    subject.fetched = datetime.datetime(2020, 2, 13)
    bibtex = subject.to_bibtex()

    file = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                        "examples",
                        "misc_orig.bib")

    assert bibtex == open(file, 'r').read()


def test_techreport(subject: BibliographicItem):
    subject.fetched = datetime.datetime(2020, 2, 13)
    subject.type = BibliographicItemType.TECHREPORT.value
    bibtex = subject.to_bibtex()

    file = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                        "examples",
                        "techreport_orig.bib")
    assert bibtex == open(file, 'r').read()


def test_manual(subject: BibliographicItem):
    subject.fetched = datetime.datetime(2020, 2, 13)
    subject.type = BibliographicItemType.MANUAL.value
    bibtex = subject.to_bibtex()

    file = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                        "examples",
                        "manual_orig.bib")

    assert bibtex == open(file, 'r').read()


def test_phdthesis(subject: BibliographicItem):
    subject.fetched = datetime.datetime(2020, 2, 13)
    subject.type = "phdthesis"
    bibtex = subject.to_bibtex()

    file = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                        "examples",
                        "phdthesis_orig.bib")

    assert bibtex == open(file, 'r').read()


def test_convert_item_to_AsciiBib(subject: BibliographicItem):
    subject.fetched = datetime.datetime(2021, 8, 30)
    adoc = subject.to_asciibib()

    file = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                        "examples",
                        "asciibib.adoc")

    assert adoc == open(file, 'r').read()


def test_initialize_with_copyright_object():
    org = Organization(
        name="Test Org",
        abbreviation="TO",
        uri="test.org"
    )
    owner = [ContributionInfo(entity=org)]
    copyright = CopyrightAssociation(owner=owner, from_="2018")
    bibitem = BibliographicItem(
      formattedref=FormattedRef(content="ISO123"),
      copyright=[copyright]
    )

    fmtref = bibitem.to_xml().find("./formattedref")

    assert fmtref is not None
    assert fmtref.text == "ISO123"
    assert fmtref.attrib["format"] == FormattedStringFormat.TEXT_PLAIN


def test_warn_invalid_type_argument_error(caplog):
    invalid = "type"

    with caplog.at_level(logging.WARNING):
        BibliographicItem(type=invalid)

    assert f"[relaton-bib] invalid document type: {invalid}" \
        in caplog.text


def test_initialise_with_owner_object():
    org = Organization(name="Test Org", abbreviation="TO", uri="test.org")
    owner = [ContributionInfo(entity=org)]
    copy = CopyrightAssociation(owner=owner, from_="2019")
    assert copy.owner == owner


def test_iconvert_item_to_BibXML_RFC(subject: BibliographicItem):
    file = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                        "examples",
                        "rfc.xml")
    tree = ET.parse(file)
    reference = tree.getroot()

    bibxml = subject.to_bibxml()

    assert elements_equal(reference, bibxml)



import dataclasses
import inspect
import pytest

import xml.etree.ElementTree as ET

from relaton_bib import StructuredIdentifier, \
    StructuredIdentifierCollection, DocumentIdType


def test_remove_data_for_cn_std():
    sid = StructuredIdentifier(
        docnumber="TEST-1999",
        type=DocumentIdType.CN_STD.value)
    sid.remove_date()
    assert sid.docnumber == "TEST"


@pytest.fixture
def subject():
    return StructuredIdentifier(
        docnumber="AGNT-007",
        type=DocumentIdType.CN_STD.value,
        agency=["agncy1", "agency2"],
        class_="class",
        partnumber="1",
        edition="2",
        version="3.4.5",
        supplementtype="sup",
        supplementnumber="6",
        language="en",
        year="1007")


def test_to_xml(subject):
    host = ET.Element("host")
    result = subject.to_xml(host)

    assert host.find("./structuredidentifier") is not None
    assert result.attrib["type"] == DocumentIdType.CN_STD.value
    assert len(result.findall("./agency")) == 2
    assert result.find("./class").text == "class"
    assert result.find("./docnumber").text == "AGNT-007"
    assert result.find("./partnumber").text == "1"
    assert result.find("./edition").text == "2"
    assert result.find("./version").text == "3.4.5"
    assert result.find("./supplementtype").text == "sup"
    assert result.find("./supplementnumber").text == "6"
    assert result.find("./language").text == "en"
    assert result.find("./year").text == "1007"


def test_to_asciibib(subject):
    result = subject.to_asciibib()

    assert result == inspect.cleandoc(
        """docnumber:: AGNT-007
           agency:: agncy1
           agency:: agency2
           type:: Chinese Standard
           class:: class
           partnumber:: 1
           edition:: 2
           version:: 3.4.5
           supplementtype:: sup
           supplementnumber:: 6
           language:: en
           year:: 1007""")


def test_to_asciibib_with_pref(subject):
    result = subject.to_asciibib(prefix="test")

    assert result == inspect.cleandoc(
        """test.docnumber:: AGNT-007
           test.agency:: agncy1
           test.agency:: agency2
           test.type:: Chinese Standard
           test.class:: class
           test.partnumber:: 1
           test.edition:: 2
           test.version:: 3.4.5
           test.supplementtype:: sup
           test.supplementnumber:: 6
           test.language:: en
           test.year:: 1007""")


def test_hash(subject):
    result = dataclasses.asdict(subject)

    assert result["type"] == DocumentIdType.CN_STD.value
    assert len(result["agency"]) == 2
    assert result["class_"] == "class"
    assert result["docnumber"] == "AGNT-007"
    assert result["partnumber"] == "1"
    assert result["edition"] == "2"
    assert result["version"] == "3.4.5"
    assert result["supplementtype"] == "sup"
    assert result["supplementnumber"] == "6"
    assert result["language"] == "en"
    assert result["year"] == "1007"


def test_collection(subject):
    col = StructuredIdentifierCollection([subject])
    assert len(col) == 1

    col.remove_part()
    assert col[0].docnumber == "AGNT"
    assert col[0].partnumber is None
    col.all_parts()
    assert col[0].docnumber == "AGNT (all parts)"



from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
from typing import List, Union

import logging
import xml.etree.ElementTree as ET

from .formatted_string import FormattedString
from .relaton_bib import lang_filter, to_ds_instance
from .person import Person
from .organization import Organization


class ContributorRoleType(str, Enum):
    AUTHOR = "author"
    PERFORMER = "performer"
    PUBLISHER = "publisher"
    EDITOR = "editor"
    ADAPTER = "adapter"
    TRANSLATOR = "translator"
    DISTRIBUTOR = "distributor"
    INSTITUTION = "institution"
    ORGANIZATION = "organization"
    SCHOOL = "school"

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_


@dataclass
class ContributorRole:
    type: str
    description: list[FormattedString] = field(default_factory=list)

    def __post_init__(self):
        if not (ContributorRoleType.has_value(self.type)):
            logging.warning(
                f"[relaton-bib] Contributor's type {self.type} is invalid")

        def str_to_formatted(s):
            return FormattedString(content=s, format=None) if s is str else s

        self.description = list(map(str_to_formatted, self.description))

    def to_xml(self, parent, opts={}):
        name = "role"
        result = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)
        result.attrib["type"] = self.type

        for d in lang_filter(self.description, opts):
            d.to_xml(ET.SubElement(result, "description"))

        return result

    def to_asciibib(self, prefix="", count=1):
        pref = f"{prefix}." if prefix else prefix
        out = [f"{prefix}::"] if count > 1 else []
        for desc in self.description:
            out.append(desc.to_asciibib(f"{pref}role.description",
                                        len(self.description)))
        if self.type:
            out.append(f"{pref}role.type:: {self.type}")
        return "\n".join(out)


@dataclass
class ContributionInfo:
    entity: Union[Person, Organization]
    role: List[ContributorRole] = field(default_factory=list)

    def __post_init__(self):
        if self.role:
            self.role = list(map(to_ds_instance(ContributorRole), self.role))
        else:
            role_type = ContributorRoleType.AUTHOR \
                if isinstance(self.entity, Person) \
                else ContributorRoleType.PUBLISHER
            self.role.append(ContributorRole(type=role_type.value))

    def to_xml(self, parent, opts={}):
        return self.entity.to_xml(parent, opts)

    def to_asciibib(self, prefix="", count=1):
        # V ported from original code but looks strange
        pref = (prefix.split(".") + [None])[0]
        out = [f"{pref}::"] if count > 1 else []
        out.append(self.entity.to_asciibib(prefix))
        for r in self.role:
            out.append(r.to_asciibib(pref, len(self.role)))
        return "\n".join(out)



import dataclasses
import pytest

import xml.etree.ElementTree as ET

from relaton_bib import DocRelationCollection, DocumentRelation, \
    FormattedRef, BibliographicItem


@pytest.fixture
def subject():
    return DocRelationCollection([
        DocumentRelation(
            type=DocumentRelation.Type.replace,
            bibitem=BibliographicItem(
                formattedref=FormattedRef(content="realtion1"),
            )),
        DocumentRelation(
            type=DocumentRelation.Type.obsoletes,
            bibitem=BibliographicItem(
                formattedref=FormattedRef(content="realtion2"),
            ))])


def test_returns_one_replace(subject):
    assert len(subject) == 2
    assert len(subject.replaces()) == 1


def test_to_xml(subject):
    host = ET.Element("host")
    result = subject.to_xml(host)

    assert len(result.findall("./relation")) == 2
    assert host.findall("./relation")[0].attrib["type"] \
        == DocumentRelation.Type.replace
    assert host.findall("./relation")[1].attrib["type"] \
        == DocumentRelation.Type.obsoletes


def test_to_asciibib(subject):
    result = subject.to_asciibib()

    assert "relation::" in result


def test_to_asciibib_with_pref(subject):
    result = subject.to_asciibib(prefix="test")

    assert "test.relation::" in result


def test_hash(subject):
    result = dataclasses.asdict(subject)

    assert result["array"][0]["type"] == DocumentRelation.Type.replace
    assert result["array"][1]["type"] == DocumentRelation.Type.obsoletes
